//----------------    DirectX ----------------
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
//     DirectX  :
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace DXApp
{
  /// <summary>
	///      DirectX
	/// </summary>
	public class Form1 : System.Windows.Forms.Form
	{
    //      Direct3D
    private Device       device   = null; //    
    private VertexBuffer trBuff   = null; //      
    private Mesh         testMesh = null; //     
    private Material     testMat;         //    
    private Mesh         tinyMesh = null; //    ,   
    private Material[]   tinyMaterials;   //    
    private Texture[]    tinyTextures;    //    
    private Microsoft.DirectX.Direct3D.Font font; //     


    //     :
    private float RotateAngX = 0; //       X;
    private float RotateAngY = 0; //       Y;
    private float RotateAngZ = 0; //       Z;
    //       :
    private Point    LastMousePt = new Point(0,0);
    private DateTime StartTime;       //    
    private long     FrameNumber = 0; // -  
    private double   FpsValue = 0;    //   FPS

		/// <summary>
		///    
		/// </summary>
		private System.ComponentModel.Container components = null;

    #region Base Form Methods

    /// <summary>
    ///   
    /// </summary>
		public Form1()
		{
			//   
			InitializeComponent();

			//    :
      this.SetStyle( ControlStyles.AllPaintingInWmPaint | //    OnPaint
                     ControlStyles.Opaque, true );    //    - 
		}

		/// <summary>
		///   -   
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
          if (device != null)
            device.Dispose();
          if (testMesh != null)
            testMesh.Dispose();
          if (trBuff != null)
            trBuff.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
      System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(Form1));
      // 
      // Form1
      // 
      this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
      this.ClientSize = new System.Drawing.Size(292, 273);
      this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
      this.Name = "Form1";
      this.Text = "DirectX Sample Application";
      this.MouseDown += new System.Windows.Forms.MouseEventHandler(this.Form1_MouseDown);
      this.Paint += new System.Windows.Forms.PaintEventHandler(this.Form1_Paint);
      this.MouseMove += new System.Windows.Forms.MouseEventHandler(this.Form1_MouseMove);

    }
		#endregion

		/// <summary>
		///     :
		/// </summary>
		[STAThread]
		static void Main() 
		{
      using (Form1 frm = new Form1())
      {
        frm.Show();
        frm.InitializeGraphics();
        Application.Run( frm );
      }
		}

    #endregion

    #region Step 1 Methods - Graphics Initialization and Rendering

    /// <summary>
    ///    :
    /// </summary>
    private void InitializeGraphics()
    {
      //    :
      PresentParameters prParam = new PresentParameters();
      prParam.Windowed = true;      //     
      prParam.SwapEffect = SwapEffect.Discard; //   
      prParam.EnableAutoDepthStencil = true; //   Z-
      prParam.AutoDepthStencilFormat = DepthFormat.D16; // Z-  16 
      prParam.MultiSample = MultiSampleType.None; //  

      //   :
      device = new Device( 0,    //     
                           DeviceType.Hardware, //   
                           this, //      
                           //     :
                           CreateFlags.SoftwareVertexProcessing,
                           prParam //    
                         );

      device.DeviceReset += new System.EventHandler( this.Device_OnReset );

      //    :
      CreateTriangleBuff();
      InitTriangleBuff();

      //    :
      testMesh = Mesh.Teapot( device );

      // C   :
      testMat = new Material();
      testMat.Ambient  = Color.LightGreen;
      testMat.Diffuse  = Color.LightGreen;
      testMat.Specular = Color.White;

      //    :
      LoadMesh(@"..\..\Meshes\tiny.x");

      //    :
      System.Drawing.Font localFont = new System.Drawing.Font( "Arial", 14.0f, FontStyle.Bold );
      font = new Microsoft.DirectX.Direct3D.Font( device, localFont );
      
      //    :
      SetLights();

      //     :
      StartTime = DateTime.Now;
    }

    /// <summary>
    ///     :
    /// </summary>
    private void Form1_Paint(object sender, System.Windows.Forms.PaintEventArgs e)
    {
      //    Fps:
      CalculateFps();
      //   :
      RenderScene();
      //  ,   :
      this.Invalidate();
    }

    /// <summary>
    ///     :
    /// </summary>
    private void RenderScene()
    {
      //    :
      device.Clear( ClearFlags.Target | ClearFlags.ZBuffer, Color.DarkBlue, 1.0f, 0 );

      //    :
      device.RenderState.CullMode = Cull.None;
      //    :
      //device.RenderState.Lighting = true;
      //     :
      //device.RenderState.NormalizeNormals = true;

      //   0-  :
      //device.Lights[0].Update();

      //    :
      //SetupMatrices();

      //    :
      device.BeginScene();

      //   :
      DrawSimpleTriangle();
      //DrawBufferedTriangle();

      //   :
      //DrawTeapotModel();

      //   :
      //DrawTinyMesh();

      //   Fps  :
      //WriteFpsText();

      //    :
      device.EndScene();

      //      :
      device.Present();
    }

    /// <summary>
    ///    :
    /// </summary>
    private void DrawSimpleTriangle()
    {
      //   :
      CustomVertex.PositionColored[] verts = new CustomVertex.PositionColored[3];
      //  A:
      verts[0].Position = new Vector3( 0.0f, 1.0f, 0.0f );
      verts[0].Color = Color.Red.ToArgb();
      //  B:
      verts[1].Position = new Vector3( 1.0f, -0.8f, 0.0f );
      verts[1].Color = Color.Green.ToArgb();
      //  C:
      verts[2].Position = new Vector3( -1.0f, -0.8f, 0.0f );
      verts[2].Color = Color.Yellow.ToArgb();

      //  :
      device.RenderState.Lighting = false;

      //  :
      device.VertexFormat = CustomVertex.PositionColored.Format;
      device.DrawUserPrimitives( PrimitiveType.TriangleList, 1, verts );
    }

    #endregion 

    #region Step 1 Additional Methods - Buffered Triangle

    /// <summary>
    ///      -   :
    /// </summary>
    private void DrawBufferedTriangle()
    {
      //  :
      device.RenderState.Lighting = false;
      
      //     :
      device.SetStreamSource( 0, trBuff, 0 );
      device.VertexFormat = CustomVertex.PositionColored.Format;
      device.DrawPrimitives( PrimitiveType.TriangleList, 0, 1 );
    }

    /// <summary>
    ///     :
    /// </summary>
    private void CreateTriangleBuff()
    {
      if (trBuff != null)
        trBuff.Dispose();

      trBuff = new VertexBuffer( typeof( CustomVertex.PositionColored ), 3, device, 
        Usage.Dynamic | Usage.WriteOnly, CustomVertex.PositionColored.Format, 
        Pool.Default );
      trBuff.Created += new EventHandler( this.trBuff_OnCreate );
    }

    /// <summary>
    ///     :
    /// </summary>
    private void trBuff_OnCreate(object sender, System.EventArgs e)
    {
      InitTriangleBuff();
    }    

    /// <summary>
    ///     :
    /// </summary>
    private void InitTriangleBuff()
    {
      if (trBuff != null)
      {
        //       :
        GraphicsStream BuffStream = trBuff.Lock(0, 0, LockFlags.None);

        //   :
        CustomVertex.PositionColored[] verts = new CustomVertex.PositionColored[3];
        //  A:
        verts[0].Position = new Vector3( 0.0f, 1.0f, 0.0f );
        verts[0].Color = Color.Red.ToArgb();
        //  B:
        verts[1].Position = new Vector3( 1.0f, -0.8f, 0.0f );
        verts[1].Color = Color.Green.ToArgb();
        //  C:
        verts[2].Position = new Vector3( -1.0f, -0.8f, 0.0f );
        verts[2].Color = Color.Yellow.ToArgb();

        //     :
        BuffStream.Write( verts );
        trBuff.Unlock();
      }
    }

    #endregion

    #region Step 2 Methods - Matrix Transformations

    /// <summary>
    ///     
    /// </summary>
    private void SetupMatrices()
    {
      //   :
      device.Transform.Projection = Matrix.PerspectiveFovLH( (float) Math.PI / 4, //    
                                                             ((float) Width) / Height, //  
                                                             0.5f,   //    
                                                             100f ); //     
      //    (   ):
      device.Transform.View = Matrix.LookAtLH( new Vector3( 0, 0.5f, -5 ),   //   
                                               new Vector3( 0, 0, 0  ),   //    
                                               new Vector3( 0, 1,  0 ) ); //    
      //   :
      device.Transform.World = Matrix.RotationX( RotateAngX ) * //    X
                               Matrix.RotationY( RotateAngY ) * //    Y
                               Matrix.RotationZ( RotateAngZ );  //    Z
    }

    /// <summary>
    ///       :
    /// </summary>
    private void Form1_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
    {
      if (e.Button == MouseButtons.Left)
      {
        //       :
        LastMousePt.X = e.X;
        LastMousePt.Y = e.Y;
      }    
    }

    /// <summary>
    ///       :
    /// </summary>
    private void Form1_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
    {
      if (e.Button == MouseButtons.Left)
      {
        //      
        float ChgX = e.X - LastMousePt.X;
        float ChgY = e.Y - LastMousePt.Y;
        //    :
        LastMousePt.X = e.X;
        LastMousePt.Y = e.Y;
        //         :
        RotateAngX += ((float) Math.PI) * ChgY / Height;
        RotateAngY -= ((float) Math.PI) * ChgX / Width;
      }    
    }

    #endregion

    #region Step 3 Methods - Rendering Mesh Object

    /// <summary>
    ///    :
    /// </summary>
    private void DrawTeapotModel()
    {
      //    :
      device.Material = testMat;
      //    :
      testMesh.DrawSubset( 0 );
    }

    /// <summary>
    ///     
    /// </summary>
    private void SetLights()
    {
      //   :
      device.Lights[0].Type = LightType.Point;
      device.Lights[0].Position = new Vector3( 5.0f, 5.0f, -5.0f );
      device.Lights[0].Diffuse = Color.White;
      device.Lights[0].Attenuation0 = 0.5f;
      device.Lights[0].Range = 100.0f;
      device.Lights[0].Enabled = true;
    }

    /// <summary>
    ///    :
    /// </summary>
    private void Device_OnReset(object sender, System.EventArgs e)
    {
      SetLights();
    }

    #endregion

    #region Step 4 Methods - Loading Mesh from X-file

    /// <summary>
    ///      X-
    /// </summary>
    private void LoadMesh( string file )
    {
      ExtendedMaterial[] mtrl;

      //    X-:
      tinyMesh = Mesh.FromFile(file, MeshFlags.Managed, device, out mtrl);

      //    ,  :
      if ((mtrl != null) && (mtrl.Length > 0))
      {
        tinyMaterials = new Material[mtrl.Length];
        tinyTextures  = new Texture[mtrl.Length];

        //     :
        for (int i = 0; i < mtrl.Length; i++)
        {
          tinyMaterials[i] = mtrl[i].Material3D;
          if ((mtrl[i].TextureFilename != null) && (mtrl[i].TextureFilename != string.Empty))
          {
            //      -   :
            tinyTextures[i] = TextureLoader.FromFile(device, @"..\..\Meshes\" + mtrl[i].TextureFilename);
          }
        }
      }
    }

    /// <summary>
    ///    ,   X-
    /// </summary>
    private void DrawTinyMesh()
    {
      //     :
      device.Transform.World *= Matrix.RotationX( -(float) Math.PI / 2 ) *
                                Matrix.RotationY(  (float) Math.PI     ) *
                                Matrix.Scaling( 0.007f, 0.007f, 0.007f );

      for (int i=0; i<tinyMaterials.Length; i++)
      {
        //    :
        device.Material = tinyMaterials[i];
        //    :
        device.SetTexture(0, tinyTextures[i]);
        //   :
        tinyMesh.DrawSubset(i);
      }
    }

    #endregion

    #region Step 5 Methods - Calculating FPS and writing Text
    
    /// <summary>
    ///     - Fps
    /// </summary>
    private void CalculateFps()
    {
      //          :
      DateTime CurrTime = DateTime.Now;
      TimeSpan TotalTime = CurrTime - StartTime;
      double TotalSeconds = TotalTime.TotalSeconds;
      FrameNumber++;
      //    - FPS:
      FpsValue = (double) FrameNumber / TotalSeconds;
    }

    /// <summary>
    ///     - Fps
    /// </summary>
    private void WriteFpsText()
    {
      string FpsText = "FPS: " + FpsValue.ToString( "G5" );
      font.DrawText(null, FpsText, new Rectangle( 0, 0, 60, 20 ), 
                    DrawTextFormat.NoClip | DrawTextFormat.ExpandTabs, Color.Yellow );
    }

    #endregion
	}
}
